/******************************************************************************
 * Author: Mario Roessler
 * Date:   2008/12/19
 * Version: 1.0
 * Description: gets characters from UART and prints information 
 *              to a 4x27 LCD Display
 *              calculates the Maidenhead locator
 * License: This program is distributed under GNU General Public License Version 2
 ******************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include "lcd_1.h"
#include "uart.h"

#define F_CPU 16000000UL  // Clock speed = 16MHz
#define XTAL 16000000

#define delay(us)  _delayFourCycles( ( ( 1*(XTAL/4000) )*us)/1000 )

#define DISP_LINE_LENGTH    27
#define UART_BAUD_RATE      4800 


/***************************************************
 * Hello World
 ***************************************************/
void hello_world(void)
{
   // Hello
   uart_puts_P("\n\rGPS2LCD v1.0\n\r");
   lcd_puts_P(1,"GPS2LCD v1.0\n");
   lcd_puts_P(2,"by Mario Roessler\n");
}


/***************************************************
 * gps_findgpgga() - search for GPGGA header
 * just searching for 3 random characters followed
 * by "GA" and returns 1 of it was found or
 * 0 if not found
 ***************************************************/
int gps_findgpgga(void)
{
   char buffers;
   uint8_t skip = 3;
   /* skip first 3 characters */
   while(skip) {
     buffers = uart_getc();
     if (buffers && UART_NO_DATA) {
	   if(PIND & (1<<PD2)) uart_putc(buffers);
	   skip--;
	   //lcd_puts_P(2,"*");
     }
   }
   
   while (1==1) {
    buffers = uart_getc();
    if (buffers && UART_NO_DATA) {
      if (buffers == 'G') {
	    if(PIND & (1<<PD2)) uart_putc(buffers);
	    //lcd_puts_P(2,"G");
	    while(1==1) {
		  buffers = uart_getc();
          if (buffers && UART_NO_DATA) {
		    if (buffers == 'A') {
			  if(PIND & (1<<PD2)) uart_putc(buffers);
			  //lcd_puts_P(2,"A");
			  return 1;
            } else return 0;
          } 
        }
      } else return 0;
    }
   }
   //lcd_puts_P(2,"-");
   return 0;
}

/***************************************************
 * skip_dataset(skip), skip is one searches for the next ','
 ***************************************************/
void skip_dataset(uint8_t skip)
{
   char buffers;
   
   /* skip datasets */
   while(skip) {
     buffers = uart_getc();
     if (buffers && UART_NO_DATA) {
	   if(PIND & (1<<PD2)) uart_putc(buffers);
	   if (buffers == ',') {
	     skip--;
	     //lcd_puts_P(2,"*");
       }
     }
   }
}

/***************************************************
 * read_data(*buffer, number_of_characters)
 ***************************************************/
void read_data(char *buffer, uint8_t num)
{
  char bufferc;
  strcpy(buffer,""); //clear buffer before reading  

  while(num) {
     bufferc = uart_getc();
     if (bufferc && UART_NO_DATA) {
	   strncat(buffer,&bufferc,1);
	   if(PIND & (1<<PD2)) uart_putc(bufferc);
	   num--;
     }
  }
}

/***************************************************
 * gps_getdata_raw() - read and print GPS data
 * data is separated by ',' - search for it
 * afterwards read:
 * time (6 characters)
 * Position n/s (9 characters)
 * North or South (1 character)
 * Position e/w (10 characters)
 * East or West (1 character)
 * number of satellites (2 characters)
 * 
 ***************************************************/
void gps_getdata_raw(void)
{
  char bufs[11];  

  skip_dataset(1); //search for start of next dataset
  read_data(bufs,6);
  lcd_gotoxy(1,0,1);
  lcd_puts_P(1,"Time: ");
  lcd_puts(1,bufs);
  
  skip_dataset(1);
  read_data(bufs,9);
  lcd_gotoxy(2,0,0);
  lcd_puts_P(2," ");
  lcd_puts(2,bufs);
  lcd_puts_P(2," ");

  skip_dataset(1);
  read_data(bufs,1);
  lcd_puts(2,bufs);

  skip_dataset(1);
  read_data(bufs,10);
  lcd_gotoxy(2,0,1);
  lcd_puts(2,bufs);
  lcd_puts_P(2," ");

  skip_dataset(1);
  read_data(bufs,1);
  lcd_puts(2,bufs);

  skip_dataset(2);
  read_data(bufs,2);
  lcd_puts_P(1,"    Sat: ");
  lcd_puts(1,bufs);

  skip_dataset(2);
  read_data(bufs,5);
  lcd_gotoxy(2,16,0);
  lcd_puts_P(2,"hNN: ");
  lcd_puts(2,bufs);
  lcd_puts_P(2,"m");

}

/***************************************************
 * cpy_str(*dest,*source)
 * copies a part of a string from source to destination
 * first searches for '.' and copies until end or ','
 ***************************************************/
uint8_t cpy_str(char *dest,char *source)
{
   int i=0;
   strncpy(dest,"\0",1);  //clear buffer
   while((source[i] != '.') && (i < strlen(source))) {
	 i++;
	 if(i == strlen(source)) return 0; //no part found after '.', nothing copied
   }
   i++; //skip the '.'
   while(i<strlen(source)) {
     if (source[i] == ',') return 1;
     strncat(dest,&source[i],1);
	 i++;
   }
   return 1;
}

/***************************************************
 * gps_getdata(*long int lat, * - if south
 *             *long int lon, * - if east
 *             *long int time,
 *             *uint8_t sat,
 *             *int hnn) 
 * - read GPS data and convert it
 * data is separated by ',' - search for it
 * afterwards read:
 * time (6 characters)
 * Position n/s (9 characters)
 * North or South (1 character)
 * Position e/w (10 characters)
 * East or West (1 character)
 * number of satellites (2 characters)
 * and print the information to the LCD display
 ***************************************************/
void gps_getdata(long int *lat,long int *lon,long int *time,uint8_t *sat,int *hnn)
{
  char bufs[7];
  char pos_lat[10],pos_ns[2],pos_lon[11],pos_ew[2],b_sat[3],b_hnn[6],b_time[7];

  /* get and print all the data */

  skip_dataset(1); //search for start of next dataset
  read_data(b_time,6);
  strcpy(bufs,b_time);
  
  skip_dataset(1);
  read_data(pos_lat,9);;

  skip_dataset(1);
  read_data(pos_ns,1);

  skip_dataset(1);
  read_data(pos_lon,10);

  skip_dataset(1);
  read_data(pos_ew,1);

  skip_dataset(2);
  read_data(b_sat,2);

  skip_dataset(2);
  read_data(b_hnn,5);

  /* now we have time to calculate the results */
  /* Time */

  *time = atol(b_time);
  
  /* now using b_time as buffer to save some memory */

  /* Latitude - Breitengrad - oben/unten */
  *lat = atol(pos_lat);
  /* copy lower part of number separated by '.' */
  cpy_str(b_time,pos_lat);
  *lat = *lat*10000;
  *lat = *lat + atol(b_time);
  if (pos_ns[0] == 'S') *lat = *lat * -1;

  /* Longitude - Lngengrad - links/rechts */
  *lon = atol(pos_lon);
  /* copy lower part of number separated by '.' */
  cpy_str(b_time,pos_lon);
  *lon = *lon*10000;
  *lon = *lon + atol(b_time);
  if (pos_ew[0] == 'W') *lon = *lon * -1;

  /* Satellites */
  *sat = (uint8_t)atoi(b_sat);

  /* Height above NN */
  *hnn = atoi(b_hnn);
  cpy_str(b_time,b_hnn);
  *hnn = *hnn*10;
  *hnn = *hnn + atoi(b_time);

}

/***************************************************
 * strins_r(char *buffer, char inschar, int num)
 * inserts a character inschar into buffer at
 * position num counted from the right side of the
 * buffer
 * attention ! this function does not check whether
 * the given buffer is big enougth to insert a
 * character ! maximum is 11 characters !
 **************************************************/
void strins_r(char *buffer, char inschar, int num)
{
  char bufstr[14];
  uint8_t i=0;

  strcpy(bufstr,buffer); //save the input
  strcpy(buffer,"\0");   //clear input
  if (num > strlen(bufstr)) return; //cannot insert before first character of string
  /* copy the characters before the character to insert */
  while(i < (strlen(bufstr)+1)) {
	if (i == (strlen(bufstr)-num)) strncat(buffer,&inschar,1); //insert the character
	strncat(buffer,&bufstr[i],1);
	i++;
  }
}

/**************************************************
 * right_arrange(uint8_t i)
 * insert leading spaces, string is ended by \0 character
 **************************************************/
void right_arrange(char *buffer, uint8_t i)
{
  // insert leading spaces, string is ended by \0 character
  i = (i - strlen(buffer)) + 1;
  while(i) {
    strins_r(buffer,' ',(int)strlen(buffer));
	i--;
  }
}

/***************************************************
 * gps_printdata (long int *lat,
 *                long int *lon,
 *   			  long int *time,
 *   			  uint8_t *sat,
 *				  int *hnn)
 * prints the given Data to LCD display  
 ***************************************************/
void gps_printdata (long int *lat,long int *lon,long int *time,uint8_t *sat,int *hnn)
{
  char strbuffer[15];

  /* Print time information */
  lcd_gotoxy(1,0,1);  
  ltoa(*time,strbuffer,10);
  strins_r(strbuffer, ':', 2);
  strins_r(strbuffer, ':', 5); // one character was already inserted
  lcd_puts_P(1,"Time:");
  right_arrange(strbuffer,9); //insert leading spaces
  lcd_puts(1,strbuffer);

  /* Latitude (Breite, oben/unten) */
  ltoa(*lat,strbuffer,10);
  lcd_gotoxy(2,0,0);
  lcd_puts_P(2,"La:");
  strins_r(strbuffer, 0xDF, 6); //insert degree character
  strins_r(strbuffer, '.', 4);
  strins_r(strbuffer, 0x27,0);  //insert minute character
  right_arrange(strbuffer,12); //insert leading spaces
  lcd_puts(2,strbuffer);
  lcd_puts_P(2," ");

  /* Longitude (Hhe, links, rechts) */
  ltoa(*lon,strbuffer,10);
  lcd_gotoxy(2,0,1);
  lcd_puts_P(2,"Lo:");
  strins_r(strbuffer, 0xDF, 6); //insert degree character
  strins_r(strbuffer, '.', 4);
  strins_r(strbuffer, 0x27,0);  //insert minute character
  right_arrange(strbuffer,12); //insert leading spaces
  lcd_puts(2,strbuffer);
  lcd_puts_P(2," ");

  /* Number of Satellites */
  itoa(*sat,strbuffer,10);
  lcd_gotoxy(1,17,1);
  lcd_puts_P(1,"S:");
  right_arrange(strbuffer,7); //insert leading spaces
  lcd_puts(1,strbuffer);

  /* Height above NN (format is decimeter !)*/
  itoa(*hnn,strbuffer,10);
  lcd_gotoxy(2,17,0);
  lcd_puts_P(2,"h:");
  strins_r(strbuffer, '.', 1);
  strncat(strbuffer,"m",1);
  right_arrange(strbuffer,7);
  lcd_puts(2,strbuffer);
}

/**************************************************
 * void nmeatodegree(&buf_lat,&buf_lon);
 * Position from GPS is delivered in format
 * ddmm.mmmm (d=degree, m=minute)
 * for calculations dd.ddddd is needed
 * this is done by this function
 **************************************************/
void nmeatodegree(unsigned long int *buf_lat, unsigned long int *buf_lon)
{  
   unsigned long int buffer;

   /* save minutes information */
   buffer = *buf_lat % 1000000;
   *buf_lat = *buf_lat - buffer; //subtract the minutes
   buffer = (buffer * 100) / 60; //scale minutes to fractional of degree
   *buf_lat = *buf_lat + buffer; //add fractional degree

   buffer = *buf_lon % 1000000;
   *buf_lon = *buf_lon - buffer; //subtract the minutes
   buffer = (buffer * 100) / 60; //scale minutes to fractional of degree
   *buf_lon = *buf_lon + buffer; //add fractional degree

}

/**************************************************
 * void calc_loc(int *lat, int *lon, char *loc)
 * calculate Maidenhead locator from coordinates
 * avaits coordinates in fractional of degree format ! not minutes !
 **************************************************/
void calc_loc(long int *lat,long int *lon, char *loc)
{
  unsigned long int buf_lat,buf_lon;
  uint8_t dig1,dig2,dig3,dig4,dig5,dig6,dig7,dig8;
  char bufc;

  strcpy(loc,"\0");
    
  buf_lat = (unsigned long int)(*lat + 90000000);  //add 90deg
  buf_lon = (unsigned long int)(*lon +180000000); //add 180deg

  /* now recalculate the position format */
  nmeatodegree(&buf_lat,&buf_lon);
  
  /* use the digit we got due to conversion to positive angles */
  buf_lat *= 10; 
  buf_lon *= 10;  

  dig1 = (uint8_t)(buf_lon/(20*10000000));
  dig1 += 1;
  dig2 = (uint8_t)(buf_lat/(10*10000000));
  dig2 += 1;

  /* First character of field */
  bufc=(char)(dig1+64);
  strncat(loc,&bufc,1);

  /* Second character of field */
  bufc=(char)(dig2+64);
  strncat(loc,&bufc,1);

  /* now get the remaining angle part for the square */
  /* and use the free digit (x10) */
  buf_lon = (buf_lon % (20*10000000))*10;
  buf_lat = (buf_lat % (10*10000000))*10;

  dig3 = (uint8_t)(buf_lon / (2*100000000));
  dig4 = (uint8_t)(buf_lat / (1*100000000));

  /* First digit of square */
  bufc=(char)(dig3+48);
  strncat(loc,&bufc,1);

  /* Second digit of square */
  bufc=(char)(dig4+48);
  strncat(loc,&bufc,1);

  /* Calculation for subsquare */
  /* longitude in fractions of 5 minutes and latitude in fractions of 2.5minutes */
  buf_lon = (buf_lon % (2*100000000))*10;
  buf_lat = (buf_lat % (1*100000000))*10;

  dig5 = (uint8_t)(buf_lon / 83333333) +1;
  dig6 = (uint8_t)(buf_lat / 41666667) +1;

  bufc=(char)(dig5+64);
  strncat(loc,&bufc,1);

  bufc=(char)(dig6+64);
  strncat(loc,&bufc,1);

  /* Calculation for extended subsquare */
  /* long in fractions of 0.5minutes and lat in fractions of 0.25minutes */
  buf_lon = (buf_lon % 83333333)*10;
  buf_lat = (buf_lat % 41666667)*10;

  dig7 = (uint8_t)(buf_lon / 83333333);
  dig8 = (uint8_t)(buf_lat / 41666667);

  /* First digit of extended subsquare */
  bufc=(char)(dig7+48);
  strncat(loc,&bufc,1);

  /* Second digit of extended subsquare */
  bufc=(char)(dig8+48);
  strncat(loc,&bufc,1);

}

/***************************************************
 * Main Program
 ***************************************************/
int main (void)
{

   char buffer;
   char strbuffer[10];
   int loop;
   char locator[9];

   strcpy(locator,"\0"); //clear locator

   /* variables to store GPS data */
   long int lat=0,lon=0,time=0;
   int hnn=0;
   uint8_t sat=0;

   strcpy(strbuffer,"-\0");

   /* Initialize UART */
   uart_init( UART_BAUD_SELECT(UART_BAUD_RATE,F_CPU) );
   sei();  // enable Interrupts

   // Initialize Display
   lcd_init(1,LCD_DISP_ON);
   lcd_init(2,LCD_DISP_ON);
   lcd_clrscr(1);
   lcd_clrscr(2);

   DDRD |= (1<<PD5);    //set PD5 to output direction
   PORTD |= (1<<PD5);   //switch LED at PD5 on
   hello_world();
   /* wait for a while */

   loop=4000;
   while(loop)
   {
      delay(0xFF);
	  loop--;
   }
   PORTD &= ~(1<<PD5);  //switch LED off

   /* clear display */
   lcd_clrscr(1);
   lcd_clrscr(2);

   /* Main Loop */
   while(1==1)
   {
      /* Search loop for $GPRMC packet */
      /* Search for start of a line = $ character */
      buffer = uart_getc();
      if (buffer && UART_NO_DATA)
      {
	    if(PIND & (1<<PD2)) uart_putc(buffer);
        if (buffer == '$') {
		   //lcd_puts_P(2,"+"); //debug: a $ sign was found
	       if (gps_findgpgga()) {
 		      //lcd_clrscr(1);
			  lcd_gotoxy(1,0,0);
		      lcd_puts_P(1,"GPGGA packet count (");
		      loop++;
		 	  itoa(loop,strbuffer,10);
			  lcd_puts(1,strbuffer);
			  lcd_puts_P(1,")");
			  //gps_getdata_raw();
			  gps_getdata(&lat,&lon,&time,&sat,&hnn);
			  gps_printdata(&lat,&lon,&time,&sat,&hnn);
			  calc_loc(&lat,&lon,locator);
              lcd_gotoxy(2,17,1);
			  lcd_puts(2,locator);
		   }
        }
      }
   
   }
}
